@@ -3,7 +3,7 @@ from argparse import ArgumentParser
 from copy import copy
 from functools import partial
 from queue import PriorityQueue
-from typing import List, Tuple, Optional, final, Dict
+from typing import List, Tuple, Optional, Dict
 
 import numpy as np
 from numpy.random import RandomState
@@ -64,7 +64,6 @@ class MdpEnvironment(Environment, ABC):
         :return: 2-tuple of next state and next reward.
         """
 
-    @final
     def run_step(
             self,
             t: int,
@@ -526,6 +525,36 @@ class PrioritizedSweepingMdpPlanningEnvironment(MdpPlanningEnvironment):
         self.bootstrap_function: Optional[partial] = None
         self.q_S_A: Optional[Dict[MdpState, Dict[Action, IncrementalSampleAverager]]] = None
 
+    def __getstate__(
+            self
+    ) -> Dict:
+        """
+        Get state dictionary for pickling.
+
+        :return: State dictionary.
+        """
+
+        state_dict = dict(self.__dict__)
+
+        # the priority queue cannot be pickled. blank it out.
+        state_dict['state_action_priority'] = None
+
+        return state_dict
+
+    def __setstate__(
+            self,
+            state
+    ):
+        """
+        Set the state dictionary.
+
+        :param state: State dictionary.
+        """
+
+        self.__dict__ = state
+
+        self.state_action_priority = PriorityQueue()
+
 
 @rl_text(chapter=8, page=174)
 class TrajectorySamplingMdpPlanningEnvironment(MdpPlanningEnvironment):
