@@ -7,6 +7,7 @@ from rlai.actions import Action
 from rlai.agents.mdp import MdpAgent
 from rlai.environments.mdp import ModelBasedMdpEnvironment
 from rlai.meta import rl_text
+from rlai.policies.tabular import TabularPolicy
 from rlai.states.mdp import MdpState
 
 
@@ -35,6 +36,9 @@ def evaluate_v_pi(
     value of delta.
     """
 
+    # noinspection PyTypeHints
+    agent.pi: TabularPolicy
+
     theta, num_iterations = check_termination_criteria(
         theta=theta,
         num_iterations=num_iterations
@@ -118,6 +122,9 @@ def evaluate_q_pi(
     (2) final value of delta.
     """
 
+    # noinspection PyTypeHints
+    agent.pi: TabularPolicy
+
     theta, num_iterations = check_termination_criteria(
         theta=theta,
         num_iterations=num_iterations
