@@ -3,12 +3,15 @@ import pickle
 from datetime import datetime
 from typing import Optional, Union
 
+from matplotlib.backends.backend_pdf import PdfPages
+
 from rlai.agents.mdp import MdpAgent
-from rlai.environments.mdp import MdpEnvironment, MdpPlanningEnvironment, PrioritizedSweepingMdpPlanningEnvironment
-from rlai.environments.openai_gym import Gym
+from rlai.environments.mdp import MdpEnvironment, MdpPlanningEnvironment
+from rlai.gpi import PolicyImprovementEvent
 from rlai.gpi.temporal_difference.evaluation import evaluate_q_pi, Mode
 from rlai.gpi.utils import plot_policy_iteration
 from rlai.meta import rl_text
+from rlai.utils import RunThreadManager
 from rlai.value_estimation import StateActionValueEstimator
 
 
@@ -22,13 +25,14 @@ def iterate_value_q_pi(
         alpha: Optional[float],
         mode: Union[Mode, str],
         n_steps: Optional[int],
-        epsilon: float,
         planning_environment: Optional[MdpPlanningEnvironment],
         make_final_policy_greedy: bool,
         q_S_A: StateActionValueEstimator,
+        thread_manager: RunThreadManager = None,
         num_improvements_per_plot: Optional[int] = None,
         num_improvements_per_checkpoint: Optional[int] = None,
-        checkpoint_path: Optional[str] = None
+        checkpoint_path: Optional[str] = None,
+        pdf_save_path: Optional[str] = None
 ):
     """
     Run temporal-difference value iteration on an agent using state-action value estimates.
@@ -43,19 +47,24 @@ def iterate_value_q_pi(
     :param alpha: Constant step size to use when updating Q-values, or None for 1/n step size.
     :param mode: Evaluation mode (see `rlai.gpi.temporal_difference.evaluation.Mode`).
     :param n_steps: Number of steps (see `rlai.gpi.temporal_difference.evaluation.evaluate_q_pi`).
-    :param epsilon: Total probability mass to spread across all actions, resulting in an epsilon-greedy policy. Must
-    be strictly > 0.
     :param planning_environment: Planning environment to learn and use.
     :param make_final_policy_greedy: Whether or not to make the agent's final policy greedy with respect to the q-values
     that have been learned, regardless of the value of epsilon used to estimate the q-values.
     :param q_S_A: State-action value estimator.
+    :param thread_manager: Thread manager. The current function (and the thread running it) will wait on this manager
+    before starting each iteration. This provides a mechanism for pausing, resuming, and aborting training. Omit for no
+    waiting.
     :param num_improvements_per_plot: Number of improvements to make before plotting the per-improvement average. Pass
     None to turn off all plotting.
     :param num_improvements_per_checkpoint: Number of improvements per checkpoint save.
     :param checkpoint_path: Checkpoint path. Must be provided if `num_improvements_per_checkpoint` is provided.
+    :param pdf_save_path: Path where a PDF of all plots is to be saved, or None for no PDF.
     """
 
-    if epsilon is None or epsilon <= 0:
+    if thread_manager is None:
+        thread_manager = RunThreadManager(True)
+
+    if q_S_A.epsilon is None or q_S_A.epsilon <= 0:
         raise ValueError('epsilon must be strictly > 0 for TD-learning')
 
     if checkpoint_path is not None:
@@ -64,6 +73,10 @@ def iterate_value_q_pi(
     if isinstance(mode, str):
         mode = Mode[mode]
 
+    pdf = None
+    if pdf_save_path is not None:
+        pdf = PdfPages(os.path.expanduser(pdf_save_path))
+
     i = 0
     iteration_average_reward = []
     iteration_total_states = []
@@ -72,6 +85,11 @@ def iterate_value_q_pi(
     start_datetime = datetime.now()
     while i < num_improvements:
 
+        thread_manager.wait()
+
+        if thread_manager.abort:
+            break
+
         print(f'Value iteration {i + 1}:  ', end='')
 
         # interact with the environment and (optionally) build a model of the environment for planning purposes
@@ -80,7 +98,6 @@ def iterate_value_q_pi(
             environment=environment,
             num_episodes=num_episodes_per_improvement,
             num_updates_per_improvement=num_updates_per_improvement,
-            epsilon=epsilon,
             alpha=alpha,
             mode=mode,
             n_steps=n_steps,
@@ -91,10 +108,13 @@ def iterate_value_q_pi(
         num_states_improved = q_S_A.improve_policy(
             agent=agent,
             states=evaluated_states,
-            epsilon=epsilon
+            event=PolicyImprovementEvent.FINISHED_EVALUATION
         )
 
-        q_S_A.plot(final=False)
+        q_S_A.plot(
+            final=i == num_improvements - 1,
+            pdf=pdf
+        )
 
         iteration_average_reward.append(average_reward)
         iteration_total_states.append(len(q_S_A))
@@ -113,13 +133,13 @@ def iterate_value_q_pi(
                 alpha=alpha,
                 mode=mode,
                 n_steps=n_steps,
-                epsilon=epsilon,
                 planning_environment=None,
                 make_final_policy_greedy=False,
                 q_S_A=q_S_A,
                 num_improvements_per_plot=None,
                 num_improvements_per_checkpoint=None,
-                checkpoint_path=None
+                checkpoint_path=None,
+                pdf_save_path=None
             )
             print('Finished planning.')
 
@@ -132,20 +152,10 @@ def iterate_value_q_pi(
         i += 1
 
         if num_improvements_per_plot is not None and i % num_improvements_per_plot == 0:
-            plot_policy_iteration(iteration_average_reward, iteration_total_states, iteration_num_states_improved, elapsed_seconds_average_rewards)
+            plot_policy_iteration(iteration_average_reward, iteration_total_states, iteration_num_states_improved, elapsed_seconds_average_rewards, pdf)
 
         if num_improvements_per_checkpoint is not None and i % num_improvements_per_checkpoint == 0:
 
-            # gym environments cannot be pickled, so just save the native id so that we can resume it later.
-            gym_native = None
-            if isinstance(environment, Gym):
-                gym_native = environment.gym_native
-                environment.gym_native = environment.gym_native.spec.id
-
-            # priority queues cannot be pickled since they contain thread locks
-            if isinstance(planning_environment, PrioritizedSweepingMdpPlanningEnvironment):
-                planning_environment.state_action_priority = None
-
             resume_args = {
                 'agent': agent,
                 'environment': environment,
@@ -155,28 +165,28 @@ def iterate_value_q_pi(
                 'alpha': alpha,
                 'mode': mode,
                 'n_steps': n_steps,
-                'epsilon': epsilon,
                 'planning_environment': planning_environment,
                 'make_final_policy_greedy': make_final_policy_greedy,
                 'q_S_A': q_S_A,
                 'num_improvements_per_plot': num_improvements_per_plot,
                 'num_improvements_per_checkpoint': num_improvements_per_checkpoint,
-                'checkpoint_path': checkpoint_path
+                'checkpoint_path': checkpoint_path,
+                'pdf_save_path': pdf_save_path
             }
 
             with open(checkpoint_path, 'wb') as checkpoint_file:
                 pickle.dump(resume_args, checkpoint_file)
 
-            if gym_native is not None:
-                environment.gym_native = gym_native
-
     print(f'Value iteration of q_pi terminated after {i} iteration(s).')
 
-    q_S_A.plot(final=True)
-
     if make_final_policy_greedy:
+        q_S_A.epsilon = 0.0
         q_S_A.improve_policy(
             agent=agent,
             states=None,
-            epsilon=0.0
+            event=PolicyImprovementEvent.MAKING_POLICY_GREEDY
         )
+
+    if pdf is not None:
+        pdf.close()
+        print(f'PDF of plots is available at {pdf_save_path}')
