@@ -1,7 +1,8 @@
 import importlib
+import threading
 from argparse import Namespace, ArgumentParser
 from importlib import import_module
-from typing import List, Any, Optional, Callable, Tuple
+from typing import List, Any, Optional, Callable, Tuple, TextIO
 
 import numpy as np
 from numpy.random import RandomState
@@ -146,6 +147,7 @@ def sample_list_item(
 
     cum_probs = probs.cumsum()
     final_cum_prob = cum_probs[-1]
+
     if abs(1.0 - final_cum_prob) > 0.0000001:
         raise ValueError(f'Expected cumulative probabilities to sum to 1, but got {final_cum_prob} instead.')
 
@@ -159,7 +161,7 @@ def sample_list_item(
 
 
 def import_function(
-        name: str
+        name: Optional[str]
 ) -> Optional[Callable]:
     """
     Import function from fully-qualified name.
@@ -195,6 +197,26 @@ def load_class(
     return class_ref
 
 
+def get_argument_parser(
+        fully_qualified_class_name
+) -> ArgumentParser:
+    """
+    Get argument parser for a class.
+
+    :param fully_qualified_class_name: Name.
+    :return: Argument parser.
+    """
+
+    parser = None
+    try:
+        loaded_class = load_class(fully_qualified_class_name)
+        parser = loaded_class.get_argument_parser()
+    except Exception:
+        pass
+
+    return parser
+
+
 def get_base_argument_parser(
         **kwargs
 ) -> ArgumentParser:
@@ -214,7 +236,7 @@ def get_base_argument_parser(
     parser.add_argument(
         '--help',
         action='store_true',
-        help='Print usage and argument descriptions.'
+        help='Pass this flag to print usage and argument descriptions.'
     )
 
     return parser
@@ -249,3 +271,81 @@ def parse_arguments(
     del parsed_args.help
 
     return parsed_args, unparsed_args
+
+
+class StdStreamTee:
+    """
+    Standard stream tee.
+    """
+
+    def write(
+            self,
+            s: str
+    ):
+        """
+        Write.
+
+        :param s: String.
+        """
+
+        if self.print_to_stream:
+            self.stream.write(s)
+
+        if s != '\n':
+            self.buffer.append(s)
+            while len(self.buffer) > self.max_buffer_len:
+                del self.buffer[0]
+
+    def flush(
+            self
+    ):
+        """
+        Flush the stream.
+        """
+
+        self.stream.flush()
+
+    def __init__(
+            self,
+            stream: TextIO,
+            max_buffer_len: int,
+            print_to_stream: bool
+    ):
+        """
+        Initialize the reader.
+
+        :param stream: Standard stream.
+        :param max_buffer_len: Maximum buffer length.
+        :param print_to_stream: Whether or not to print back to stream.
+        """
+
+        self.stream = stream
+        self.max_buffer_len = max_buffer_len
+        self.print_to_stream = print_to_stream
+
+        self.buffer = []
+
+
+class RunThreadManager(threading.Event):
+    """
+    Manager for run threads, combining a wait event with an abort signal.
+    """
+
+    def __init__(
+            self,
+            initial_flag: bool
+    ):
+        """
+        Initialize the thread manager.
+
+        :param initial_flag: Initial flag value (see threading.Thread for the use of this argument).
+        """
+
+        super().__init__()
+
+        if initial_flag:
+            self.set()
+        else:
+            self.clear()
+
+        self.abort = False
