@@ -2,11 +2,14 @@ from abc import ABC, abstractmethod
 from argparse import ArgumentParser
 from typing import Optional, Iterable, Tuple, List, Any, Iterator
 
+import matplotlib.pyplot as plt
+from matplotlib.backends.backend_pdf import PdfPages
 from numpy.random import RandomState
 
 from rlai.actions import Action
 from rlai.agents.mdp import MdpAgent
 from rlai.environments.mdp import MdpEnvironment
+from rlai.gpi import PolicyImprovementEvent
 from rlai.meta import rl_text
 from rlai.policies import Policy
 from rlai.states.mdp import MdpState
@@ -31,7 +34,6 @@ class ValueEstimator(ABC):
         :param value: New value.
         :param weight: Weight.
         """
-        pass
 
     @abstractmethod
     def get_value(
@@ -42,7 +44,6 @@ class ValueEstimator(ABC):
 
         :return: Value.
         """
-        pass
 
     def __str__(
             self
@@ -73,7 +74,6 @@ class ActionValueEstimator(ABC):
         :param action: Action.
         :return: Value estimator.
         """
-        pass
 
     @abstractmethod
     def __len__(
@@ -84,7 +84,6 @@ class ActionValueEstimator(ABC):
 
         :return: Number of actions.
         """
-        pass
 
     @abstractmethod
     def __iter__(
@@ -95,7 +94,6 @@ class ActionValueEstimator(ABC):
 
         :return: Iterator.
         """
-        pass
 
     @abstractmethod
     def __contains__(
@@ -108,7 +106,6 @@ class ActionValueEstimator(ABC):
         :param action: Action.
         :return: True if defined and False otherwise.
         """
-        pass
 
 
 @rl_text(chapter='Value Estimation', page=23)
@@ -127,7 +124,15 @@ class StateActionValueEstimator(ABC):
         :return: Argument parser.
         """
 
-        return get_base_argument_parser()
+        parser = get_base_argument_parser()
+
+        parser.add_argument(
+            '--epsilon',
+            type=float,
+            help='Total probability mass to spread across all actions. Omit or pass 0.0 to produce a purely greedy agent.'
+        )
+
+        return parser
 
     @classmethod
     @abstractmethod
@@ -135,8 +140,7 @@ class StateActionValueEstimator(ABC):
             cls,
             args: List[str],
             random_state: RandomState,
-            environment: MdpEnvironment,
-            epsilon: float
+            environment: MdpEnvironment
     ) -> Tuple[Any, List[str]]:
         """
         Initialize a state-action value estimator from arguments.
@@ -144,10 +148,8 @@ class StateActionValueEstimator(ABC):
         :param args: Arguments.
         :param random_state: Random state.
         :param environment: Environment.
-        :param epsilon: Epsilon.
         :return: 2-tuple of a state-action value estimator and a list of unparsed arguments.
         """
-        pass
 
     @abstractmethod
     def get_initial_policy(
@@ -158,7 +160,6 @@ class StateActionValueEstimator(ABC):
 
         :return: Policy.
         """
-        pass
 
     def initialize(
             self,
@@ -176,35 +177,55 @@ class StateActionValueEstimator(ABC):
         :param weighted: Whether the estimator should be weighted.
         :return:
         """
-        pass
 
     @abstractmethod
     def improve_policy(
             self,
             agent: MdpAgent,
             states: Optional[Iterable[MdpState]],
-            epsilon: float
-    ) -> int:
+            event: PolicyImprovementEvent
+    ):
         """
         Improve an agent's policy using the current state-action value estimates.
 
         :param agent: Agent whose policy should be improved.
         :param states: States to improve, or None for all states.
-        :param epsilon: Epsilon.
+        :param event: Event that triggered the improvement.
         :return: Number of states improved.
         """
-        pass
+
+        if event == PolicyImprovementEvent.FINISHED_EVALUATION:
+            self.evaluation_policy_improvement_count += 1
+        elif event == PolicyImprovementEvent.UPDATED_VALUE_ESTIMATE:
+            self.value_estimate_policy_improvement_count += 1
 
     def plot(
             self,
-            final: bool
-    ):
+            final: bool,
+            pdf: Optional[PdfPages]
+    ) -> Optional[plt.Figure]:
         """
-        Plot the estimator.
+        Plot the estimator. If called from the main thread, then the rendering schedule will be checked and a new plot
+        will be generated per the schedule. If called from a background thread, then the data used by the plot will be
+        updated but a plot will not be generated or updated. This supports a pattern in which a background thread
+        generates new plot data, and a UI thread (e.g., in a Jupyter notebook) periodically calls `update_plot` to
+        redraw the plot with the latest data.
 
         :param final: Whether or not this is the final time plot will be called.
+        :param pdf: PDF for plots, or None for no PDF.
+        :return: Matplotlib figure, if one was generated and not plotting to PDF.
+        """
+
+    def update_plot(
+            self,
+            time_step_detail_iteration: Optional[int]
+    ):
+        """
+        Update the plot of the estimator. Can only be called from the main thread.
+
+        :param time_step_detail_iteration: Iteration for which to plot time-step-level detail, or None for no detail.
+        Passing -1 will plot detail for the most recently completed iteration.
         """
-        pass
 
     def __init__(
             self,
@@ -220,10 +241,14 @@ class StateActionValueEstimator(ABC):
 
         if epsilon is None:
             epsilon = 0.0
+        elif epsilon < 0.0:
+            raise ValueError('epsilon must be >= 0')
 
         self.epsilon = epsilon
 
         self.update_count = 0
+        self.evaluation_policy_improvement_count: int = 0
+        self.value_estimate_policy_improvement_count: int = 0
 
     @abstractmethod
     def __getitem__(
@@ -236,7 +261,6 @@ class StateActionValueEstimator(ABC):
         :param state: State.
         :return: Action-value estimator.
         """
-        pass
 
     @abstractmethod
     def __len__(
@@ -247,7 +271,6 @@ class StateActionValueEstimator(ABC):
 
         :return: Number of states.
         """
-        pass
 
     @abstractmethod
     def __contains__(
@@ -260,7 +283,6 @@ class StateActionValueEstimator(ABC):
         :param state: State.
         :return: True if defined and False otherise.
         """
-        pass
 
     @abstractmethod
     def __eq__(
@@ -273,7 +295,6 @@ class StateActionValueEstimator(ABC):
         :param other: Other estimator.
         :return: True if equal and False otherwise.
         """
-        pass
 
     @abstractmethod
     def __ne__(
@@ -286,4 +307,3 @@ class StateActionValueEstimator(ABC):
         :param other: Other estimator.
         :return: True if not equal and False otherwise.
         """
-        pass
