@@ -1,26 +1,24 @@
-import math
 from argparse import ArgumentParser
 from typing import Optional, List, Tuple, Iterator, Set
 
 import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
+from matplotlib.backends.backend_pdf import PdfPages
 from numpy.random import RandomState
 from patsy.highlevel import dmatrix
 
 from rlai.actions import Action
 from rlai.agents.mdp import MdpAgent
 from rlai.environments.mdp import MdpEnvironment
+from rlai.gpi import PolicyImprovementEvent
 from rlai.meta import rl_text
 from rlai.policies.function_approximation import FunctionApproximationPolicy
 from rlai.states.mdp import MdpState
 from rlai.utils import load_class, parse_arguments
 from rlai.value_estimation import ValueEstimator, ActionValueEstimator, StateActionValueEstimator
 from rlai.value_estimation.function_approximation.models import FunctionApproximationModel
-from rlai.value_estimation.function_approximation.models.feature_extraction import (
-    FeatureExtractor,
-    StateActionInteractionFeatureExtractor
-)
+from rlai.value_estimation.function_approximation.models.feature_extraction import FeatureExtractor
 
 
 @rl_text(chapter='Value Estimation', page=195)
@@ -189,7 +187,7 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
         parser.add_argument(
             '--plot-model',
             action='store_true',
-            help='Whether or not to plot the model (e.g., coefficients).'
+            help='Pass this flag to plot the model (e.g., coefficients, diagnostics, etc.).'
         )
 
         parser.add_argument(
@@ -211,8 +209,7 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
             cls,
             args: List[str],
             random_state: RandomState,
-            environment: MdpEnvironment,
-            epsilon: Optional[float]
+            environment: MdpEnvironment
     ) -> Tuple[StateActionValueEstimator, List[str]]:
         """
         Initialize a state-action value estimator from arguments.
@@ -220,7 +217,6 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
         :param args: Arguments.
         :param random_state: Random state.
         :param environment: Environment.
-        :param epsilon: Epsilon.
         :return: 2-tuple of a state-action value estimator and a list of unparsed arguments.
         """
 
@@ -239,7 +235,6 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
 
         estimator = ApproximateStateActionValueEstimator(
             environment=environment,
-            epsilon=epsilon,
             model=model,
             feature_extractor=fex,
             **vars(parsed_args)
@@ -291,21 +286,22 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
             self,
             agent: MdpAgent,
             states: Optional[Set[MdpState]],
-            epsilon: Optional[float]
+            event: PolicyImprovementEvent
     ) -> int:
         """
         Improve an agent's policy using the current sample of experience collected through calls to `add_sample`.
 
         :param agent: Agent whose policy should be improved.
         :param states: States to improve, or None for all states.
-        :param epsilon: Epsilon.
+        :param event: Event that triggered the improvement.
         :return: Number of states improved.
         """
 
-        if epsilon is None:
-            epsilon = 0.0
-
-        self.epsilon = epsilon
+        super().improve_policy(
+            agent=agent,
+            states=states,
+            event=event
+        )
 
         # if we have pending experience, then fit the model and reset the data.
         if self.experience_pending:
@@ -319,11 +315,6 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
             self.weights = None
             self.experience_pending = False
 
-        if self.policy_improvement_count is None:
-            self.policy_improvement_count = 1
-        else:
-            self.policy_improvement_count += 1
-
         return -1 if states is None else len(states)
 
     def evaluate(
@@ -361,14 +352,12 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
 
         X = self.feature_extractor.extract(states, actions, for_fitting)
 
-        # if no formula, then use result directly.
+        # if no formula, then the feature extraction result must be a numpy.ndarray to be used directly.
         if self.formula is None:
-            if isinstance(X, pd.DataFrame):
-                X = X.to_numpy()
-            elif not isinstance(X, np.ndarray):
+            if not isinstance(X, np.ndarray):  # pragma no cover
                 raise ValueError('Expected feature extractor to return a numpy.ndarray if not a pandas.DataFrame')
 
-        # use formula with dataframe only
+        # formulas only work with dataframes
         elif isinstance(X, pd.DataFrame):
             X = dmatrix(self.formula, X)
 
@@ -380,80 +369,45 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
 
     def plot(
             self,
-            final: bool
-    ):
+            final: bool,
+            pdf: Optional[PdfPages]
+    ) -> Optional[plt.Figure]:
         """
-        Plot the estimator.
+        Plot the estimator. If called from the main thread, then the rendering schedule will be checked and a new plot
+        will be generated per the schedule. If called from a background thread, then the data used by the plot will be
+        updated but a plot will not be generated or updated. This supports a pattern in which a background thread
+        generates new plot data, and a UI thread (e.g., in a Jupyter notebook) periodically calls `update_plot` to
+        redraw the plot with the latest data.
 
         :param final: Whether or not this is the final time plot will be called.
+        :param pdf: PDF for plots, or None for no PDF.
+        :return: Matplotlib figure, if one was generated and not plotting to PDF.
         """
 
         if self.plot_model:
 
-            model_summary = self.model.get_summary(self.feature_extractor)
+            render = final or (self.plot_model_per_improvements is not None and self.evaluation_policy_improvement_count % self.plot_model_per_improvements == 0)
 
-            if 'n' in model_summary.columns:
-                raise ValueError('Feature extractor returned disallowed column:  n')
+            return self.model.plot(
+                feature_extractor=self.feature_extractor,
+                policy_improvement_count=self.evaluation_policy_improvement_count,
+                num_improvement_bins=self.plot_model_bins,
+                render=render,
+                pdf=pdf
+            )
 
-            if 'bin' in model_summary.columns:
-                raise ValueError('Feature extractor returned disallowed column:  bin')
+    def update_plot(
+            self,
+            time_step_detail_iteration: Optional[int]
+    ):
+        """
+        Update the plot of the estimator. Can only be called from the main thread.
 
-            model_summary['n'] = self.policy_improvement_count - 1
+        :param time_step_detail_iteration: Iteration for which to plot time-step-level detail, or None for no detail.
+        Passing -1 will plot detail for the most recently completed iteration.
+        """
 
-            if self.plot_df is None:
-                self.plot_df = model_summary
-            else:
-                self.plot_df = self.plot_df.append(model_summary, ignore_index=True)
-
-            if final or (self.plot_model_per_improvements is not None and self.policy_improvement_count % self.plot_model_per_improvements == 0):
-
-                if self.plot_model_bins is None:
-                    improvements_per_bin = 1
-                    self.plot_df['bin'] = self.plot_df.n
-                else:
-                    improvements_per_bin = math.ceil(self.policy_improvement_count / self.plot_model_bins)
-                    self.plot_df['bin'] = [int(n / improvements_per_bin) for n in self.plot_df.n]
-
-                if isinstance(self.feature_extractor, StateActionInteractionFeatureExtractor):
-
-                    # set up plots
-                    feature_names = self.plot_df.feature_name.unique().tolist()
-                    n_rows = len(feature_names)
-                    action_names = [a.name for a in self.feature_extractor.actions]
-                    n_cols = len(action_names)
-                    fig, axs = plt.subplots(
-                        nrows=n_rows,
-                        ncols=n_cols,
-                        sharex='all',
-                        sharey='row',
-                        figsize=(3 * n_cols, 3 * n_rows)
-                    )
-
-                    # plot one row per feature, with actions as the columns
-                    for i, feature_name in enumerate(feature_names):
-                        feature_df = self.plot_df[self.plot_df.feature_name == feature_name]
-                        boxplot_axs = axs[i, :]
-                        feature_df.boxplot(column=action_names, by='bin', ax=boxplot_axs)
-                        boxplot_axs[0].set_ylabel(f'w({feature_name})')
-
-                    # reset labels and titles
-                    for i, row in enumerate(axs):
-                        for ax in row:
-
-                            if i < axs.shape[0] - 1:
-                                ax.set_xlabel('')
-                            else:
-                                ax.set_xlabel('Iteration' if self.plot_model_bins is None else f'Bin of {improvements_per_bin} iterations')
-
-                            if i > 0:
-                                ax.set_title('')
-
-                    fig.suptitle('Model coefficients over iterations')
-                    plt.tight_layout()
-                    plt.show()
-
-                else:
-                    raise ValueError(f'Unknown feature extractor type:  {type(self.feature_extractor)}')
+        self.model.update_plot(time_step_detail_iteration)
 
     def __init__(
             self,
@@ -462,7 +416,7 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
             model: FunctionApproximationModel,
             feature_extractor: FeatureExtractor,
             formula: Optional[str],
-            plot_model: Optional[bool],
+            plot_model: bool,
             plot_model_per_improvements: Optional[int],
             plot_model_bins: Optional[int]
     ):
@@ -489,8 +443,9 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
         overhead with each call to the formula parser. A faster alternative is to avoid formula specification (pass
         None here) and return the feature matrix directly from the feature extractor as a numpy.ndarray.
         :param plot_model: Whether or not to plot the model.
-        :param plot_model_per_improvements: Number of policy improvements between between plots.
-        :param plot_model_bins: Number of plotting bins, or None for no binning.
+        :param plot_model_per_improvements: Number of policy improvements between between plots of the model. Only used
+        if `plot_model` is True. Pass None to only plot the model at the end.
+        :param plot_model_bins: Number of plotting bins. Only used if `plot_model` is True. Pass None for no binning.
         """
 
         super().__init__(
@@ -510,8 +465,6 @@ class ApproximateStateActionValueEstimator(StateActionValueEstimator):
         self.experience_values: List[float] = []
         self.weights: np.ndarray = None
         self.experience_pending: bool = False
-        self.policy_improvement_count: Optional[int] = None
-        self.plot_df: Optional[pd.DataFrame] = None
 
     def __getitem__(
             self,
