@@ -6,6 +6,7 @@ from numpy.random import RandomState
 from rlai.actions import Action
 from rlai.agents.mdp import MdpAgent
 from rlai.environments.mdp import MdpEnvironment
+from rlai.gpi import PolicyImprovementEvent
 from rlai.gpi.improvement import improve_policy_with_q_pi
 from rlai.meta import rl_text
 from rlai.policies.tabular import TabularPolicy
@@ -247,8 +248,7 @@ class TabularStateActionValueEstimator(StateActionValueEstimator):
             cls,
             args: List[str],
             random_state: RandomState,
-            environment: MdpEnvironment,
-            epsilon: float
+            environment: MdpEnvironment
     ) -> Tuple[StateActionValueEstimator, List[str]]:
         """
         Initialize a state-action value estimator from arguments.
@@ -256,7 +256,6 @@ class TabularStateActionValueEstimator(StateActionValueEstimator):
         :param args: Arguments.
         :param random_state: Random state.
         :param environment: Environment.
-        :param epsilon: Epsilon.
         :return: 2-tuple of a state-action value estimator and a list of unparsed arguments.
         """
 
@@ -264,7 +263,6 @@ class TabularStateActionValueEstimator(StateActionValueEstimator):
 
         estimator = TabularStateActionValueEstimator(
             environment=environment,
-            epsilon=epsilon,
             **vars(parsed_args)
         )
 
@@ -310,21 +308,22 @@ class TabularStateActionValueEstimator(StateActionValueEstimator):
             self,
             agent: MdpAgent,
             states: Optional[Iterable[MdpState]],
-            epsilon: float
+            event: PolicyImprovementEvent
     ) -> int:
         """
         Improve an agent's policy using the current state-action value estimates.
 
         :param agent: Agent whose policy should be improved.
         :param states: States to improve, or None for all states.
-        :param epsilon: Epsilon.
+        :param event: Event that triggered the improvement.
         :return: Number of states improved.
         """
 
-        if epsilon is None:
-            epsilon = 0.0
-
-        self.epsilon = epsilon
+        super().improve_policy(
+            agent=agent,
+            states=states,
+            event=event
+        )
 
         q_pi = {
             s: {
